use std::{path::Path, str::FromStr, collections::HashSet, ops::Div, cmp::{min, max}};

use itertools::Itertools;
use num::Integer;

use crate::{error::AdventError, input::read_into};

struct Chart {
    stride: usize,
    space: Vec<char>,
    empty_rows: HashSet<usize>,
    empty_columns: HashSet<usize>
}

impl FromStr for Chart {
    type Err = AdventError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let stride = s.lines().next().ok_or(AdventError("No first line".into()))?.len();
        Ok(Chart {
            stride,
            space: s.chars().filter(|c| !c.is_ascii_whitespace()).collect_vec(),
            empty_rows: HashSet::<usize>::new(),
            empty_columns: HashSet::<usize>::new(),
        })
    }
}

impl Chart {
    fn detect_empties(&mut self) {
        for x in 0..self.stride {
            if self.space.iter().skip(x).step_by(self.stride).all(|c| *c == '.') {
                self.empty_columns.insert(x);
            }
        }
        for y in 0..self.space.len().div(self.stride) {
            if self.space.iter().skip(y*self.stride).take(self.stride).all(|c| *c == '.') {
                self.empty_rows.insert(y);
            }
        }
    }

    fn find_galaxys(&self) -> Vec<(usize, usize)> {
        let mut vec = vec![];
        for (i, c) in self.space.iter().enumerate() {
            if *c == '#' {
                let x = i.div_mod_floor(&self.stride);
                vec.push((x.1, x.0));
            }
        }
        vec
    }

    fn distance(&self, c1: &(usize, usize), c2: &(usize, usize), empty_distance: u64) -> u64 {
        let mut steps = 0;
        for x in (min(c1.0, c2.0)..=max(c1.0, c2.0)).into_iter().skip(1) {
            if self.empty_columns.contains(&x) {
                steps += empty_distance;
            } else {
                steps += 1;
            }
        }
        for y in (min(c1.1, c2.1)..=max(c1.1, c2.1)).into_iter().skip(1) {
            if (self.empty_rows).contains(&y) {
                steps += empty_distance;
            } else {
                steps += 1;
            }
        }
        steps
    }
}

pub fn solve(empty_distance: u64) -> Result<u64, AdventError> {
    let mut chart = read_into::<Chart>(Path::new("resources/input11.txt"))?;
    chart.detect_empties();
    let galaxy_coords = chart.find_galaxys();
    let mut sum: u64 = 0;
    for (i, c1) in galaxy_coords.iter().enumerate() {
        for c2 in galaxy_coords.iter().skip(i+1) {
            sum += chart.distance(c1, c2, empty_distance);
        }
    }
    Ok(sum)
}

pub fn one() -> Result<u64, AdventError> {
    solve(2)
}

pub fn two() -> Result<u64, AdventError> {
    solve(1000000)
}